#!/bin/bash
set -eu

# read flag
FLAG=$(<./flag.txt)

PLAYER_INFO_PATH=/tmp/player_info.json
ACCOUNTS_PATH=/tmp/accounts.json
# generate two accounts
#  [0] deployer
#  [1] player
cast wallet new --number=2 --json > "${ACCOUNTS_PATH}"
DEPLOYER_ADDRESS=$(jq -r ".[0].address" < "${ACCOUNTS_PATH}")
DEPLOYER_PRIVATE_KEY=$(jq -r ".[0].private_key" < "${ACCOUNTS_PATH}")

PLAYER_ADDRESS=$(jq -r ".[1].address" < "${ACCOUNTS_PATH}")
PLAYER_PRIVATE_KEY=$(jq -r ".[1].private_key" < "${ACCOUNTS_PATH}")

echo "DEPLOYER: ${DEPLOYER_ADDRESS}"
echo "PLAYER  : ${PLAYER_ADDRESS}"

geth --dev \
  --http --http.api="eth,web3,net" --http.vhosts '*' --http.addr "0.0.0.0" &
GETH_PID=$!

function cleanup {
  kill $GETH_PID
}

trap cleanup EXIT

# wait for geth to come up
for i in {1..5}
do
  sleep 2
  if cast block-number > /dev/null ; then
    break
  fi
done

# fund the accounts
FUND_JS="const tx1 = eth.sendTransaction({from: eth.accounts[0], to: '${DEPLOYER_ADDRESS}', value: web3.toWei(50, 'ether')})"
FUND_JS="${FUND_JS}; const tx2 = eth.sendTransaction({from: eth.accounts[0], to: '${PLAYER_ADDRESS}', value: web3.toWei(1, 'ether')})"
geth attach --exec "${FUND_JS}" /tmp/geth.ipc

# ensure that the accounts were funded
SUCCESS=0
for i in {1..5}
do
  echo "balance check $i"
  sleep 2
  BALANCE1=$(cast balance "${DEPLOYER_ADDRESS}")
  BALANCE2=$(cast balance "${PLAYER_ADDRESS}")
  if [ "$BALANCE1" != "0" ] && [ "$BALANCE2" != "0" ] ; then
    SUCCESS=1
    break
  fi
done

if [ "$SUCCESS" != "1" ] ; then
  echo "Could not confirm balance of deployer"
  exit 1
fi

DEPLOYMENT_INFO_PATH=/tmp/deployment.json
# perform the deployment
forge create src/Setup.sol:Setup --private-key "${DEPLOYER_PRIVATE_KEY}" --json > "${DEPLOYMENT_INFO_PATH}"

SETUP_ADDR=$(jq -r '.deployedTo' < "${DEPLOYMENT_INFO_PATH}")
jq --null-input --arg SETUP "${SETUP_ADDR}" --arg ADDR "${PLAYER_ADDRESS}" --arg KEY "${PLAYER_PRIVATE_KEY}" \
  '. + {"setup": $SETUP, "address": $ADDR, "private_key": $KEY}' > "${PLAYER_INFO_PATH}"

# serve the player info on port 1337
function serve_player_info {
  while true
  do
    nc -l -p 1337 -q 0 -N < "${PLAYER_INFO_PATH}"
  done
}

serve_player_info &
echo "Deployment done, capture some flags"

# if the challenge is solved, serve the flag on port 1234
while :
do
  IS_SOLVED=$(cast call "${SETUP_ADDR}" "isSolved()(bool)")
  echo "${IS_SOLVED}"
  if [[ "${IS_SOLVED}" == "true" ]] ; then
    echo "${FLAG}" | nc -l -p 1234 -q 1 -N
    exit 0
  fi
  sleep 2
done

exit 1
